// "Basic File IO" Header
#include "MemAccess.h"

// CH376S Specific stuff
#define ANSW_RET_SUCCESS		0x51	// Operation successful

#define ANSW_USB_INT_SUCCESS	0x14	// Operation successful, no further data
#define ANSW_USB_INT_CONNECT	0x15	// New USB device connected
#define ANSW_USB_INT_DISCONNECT	0x16	// USB device unplugged

#define ANSW_USB_INT_USB_READY	0x18	// Device is ready
#define ANSW_USB_INT_DISK_READ	0x1D	// Disk read operation
#define ANSW_USB_INT_DISK_WRITE	0x1E	// Disk write operation

#define ANSW_RET_ABORT			0x5F	// Operation failure
#define ANSW_USB_INT_DISK_ERR	0x1F	// USB storage device error
#define ANSW_USB_INT_BUF_OVER	0x17	// Buffer overflow
#define ANSW_ERR_OPEN_DIR		0x41	// Tried to open a directory with FILE_OPEN
#define ANSW_ERR_MISS_FILE		0x42	// File not found
#define ANSW_ERR_FOUND_NAME		0x43
#define ANSW_ERR_DISK_DISCON	0x82	// Disk disconnected
#define ANSW_ERR_LARGE_SECTOR	0x84	// Sector size is not 512 bytes
#define ANSW_ERR_TYPE_ERROR		0x92	// Invalid partition type, reformat drive
#define ANSW_ERR_BPB_ERROR		0xA1	// Partition not formatted
#define ANSW_ERR_DISK_FULL		0xB1	// Disk full
#define ANSW_ERR_FDT_OVER		0xB2	// Directory full
#define ANSW_ERR_FILE_CLOSE		0xB4	// Attempted operation on closed file

#define CH376_ERR_OVERFLOW		0x03
#define CH376_ERR_TIMEOUT		0x02
#define CH376_ERR_NO_RESPONSE	0x01
#define CH376_ERR_LONGFILENAME	0x04

#define CH376_ATTR_READ_ONLY	0x01	// Read-only file
#define CH376_ATTR_HIDDEN		0x02	// Hidden file
#define CH376_ATTR_SYSTEM		0x04	// System file
#define CH376_ATTR_VOLUME_ID	0x08	// Volume label
#define CH376_ATTR_DIRECTORY	0x10	// Subdirectory (folder)
#define CH376_ATTR_ARCHIVE		0x20	// Archive (normal) file



#define GET_IC_VER				0x01
#define SET_BAUDRATE			0x02
#define ENTER_SLEEP				0x03

#define RESET_ALL				0x05
#define CHECK_EXIST				0x06

#define SET_SDO_INT				0x0B
#define GET_FILE_SIZE			0x0C

#define SET_USB_MODE			0x15

#define GET_STATUS				0x22
#define RD_USB_DATA0			0x27

#define WR_USB_DATA				0x2C
#define WR_REQ_DATA				0x2D
#define WR_OFS_DATA				0x2E
#define SET_FILE_NAME			0x2F
#define DISK_CONNECT			0x30
#define DISK_MOUNT				0x31
#define FILE_OPEN				0x32
#define FILE_ENUM_GO			0x33
#define FILE_CREATE				0x34
#define FILE_ERASE				0x35
#define FILE_CLOSE				0x36
#define DIR_INFO_READ			0x37
#define DIR_INFO_SAVE			0x38
#define BYTE_LOCATE				0x39
#define BYTE_READ				0x3A
#define BYTE_RD_GO				0x3B
#define BYTE_WRITE				0x3C
#define BYTE_WR_GO				0x3D
#define DISK_CAPACITY			0x3E
#define DISK_QUERY				0x3F
#define DIR_CREATE				0x40

#define SEG_LOCATE				0x4A
#define SEC_READ				0x4B
#define SEC_WRITE				0x4C

#define DISK_BOC_CMD			0x50

#define DISK_READ				0x54
#define DISK_RD_GO				0x55
#define DISK_WRITE				0x56
#define DISK_WR_GO				0x57

#define MODE_HOST_INV			0x04
#define MODE_HOST_0				0x05
#define MODE_HOST_1				0x07
#define MODE_HOST_2				0x06
#define MODE_HOST_SD			0x03
#define MODE_DEFAULT			0x00

#define CMD_RET_SUCCESS			0x51
#define CMD_RET_ABORT			0x5F
#define USB_INT_USB_READY		0x18	// Device is ready
#define USB_INT_DISK_READ		0x1D	// Disk read operation
#define USB_INT_DISK_WRITE		0x1E	// Disk write operation




#define _SPI_INT				0b01000000

#define SPI_read()				SPI_transfer(0xFF)

//#define DEBUG
#ifdef DEBUG
	#define dbg_printf(...)		printf(__VA_ARGS__)
#else
	#define dbg_printf(...)
#endif

// Directory reading states
enum{
	ST_REQ,
	ST_READ,
	ST_DONE
};

uint8_t _ch376_error = 0;
uint8_t _ch376_answer = 0;
uint8_t _ch376_deviceAttached = 0;

uint8_t __DR_STATE__ = ST_REQ;			// Directory reading state machine
uint8_t __DR_ANSWER__ = 0;				// Directory reading answer

void delay10(uint16_t tms);
void setVIA(void);
void writeSS0(bool value);
void writeSS1(bool value);
uint8_t SPI_transfer(uint8_t value);

bool checkIntMessage(void);
uint8_t getInterrupt(void);
uint8_t waitInterrupt(void);

bool initDevice(void);
uint8_t mount(void);

void driveAttach(void);
void driveDetach(void);

bool pingDevice(void);
#define getDriveStatus()		(_ch376_deviceAttached)
#define checkAnswer()			(_ch376_answer)

uint8_t openPath(char *path, bool open, bool chcwd);






// Mostly hardware independent functions and variables

#define O_BINARY			(0b00000001)	// Binary Mode (ie no automatic LF/CR conversion)
#define O_APPEND			(0b00000010)	// Apend Mode (ie start at the end of the file)
#define O_CREATE			(0b00000100)	// Create if not exist (ie if a file is not found, create it)
#define O_TRUNC				(0b00001000)	// Truncate (ie if the file already exists, delete it's contentd)
#define O_WRITE				(0b00010000)	// Write Mode (ie update file attributes when closing)

#define A_READ_ONLY			(0b00000001)
#define A_HIDDEN			(0b00000010)
#define A_SYSTEM			(0b00000100)
#define A_VOLUME_ID			(0b00001000)
#define A_DIRECTORY			(0b00010000)
#define A_ARCHIVE			(0b00100000)
#define A_LONG_NAME			(ATTR_READ_ONLY | ATTR_HIDDEN | ATTR_SYSTEM | ATTR_VOLUME_ID)
#define A_LONG_NAME_MASK	(ATTR_LONG_NAME | ATTR_DIRECTORY | ATTR_ARCHIVE)

enum{
	F_SUCCESS,				// General Success
	F_ERROR,				// General Error
	F_STOP_EOF,				// Stop, because: "End of File" character
	F_STOP_EOD,				// Stop, because: "End of Directory"
	F_STOP_NL,				// Stop, because: "New Line" character
	F_FILE_NOT_FOUND,		// Error, File not found
	F_DIR_NOT_FOUND,		// Error, Directory not found
	F_PATH_TOO_LONG,		// Error, Path too long
	F_NOT_WRITABLE,			// Error, Tried writing to a read-only file
};


typedef struct __file_struct{
	uint32_t fptr;			// (4) File Pointer
	uint32_t fsize;			// (4) File Size
	char *fpath;			// (4) Pointer to the full Path of the File
	uint16_t fcrd;			// (2) Creation Date
	uint16_t flmd;			// (2) Last Modified Date
	uint8_t ferror;			// (1) Error Code
	uint8_t fmode;			// (1) File Mode
} file;						// 18 Bytes in total (or 16 when fpath is using a 16-bit Pointer)


typedef struct __dir_struct {
	char fname[11];			// (11) File Name
	uint8_t fattr;			// ( 1) File Attributes
	uint8_t fuattr;			// ( 1) User Attributes
	uint8_t fctimet;		// ( 1) Creation Time 1/10th seconds
	uint16_t fctimes;		// ( 2) Creation Time
	uint16_t fcdate;		// ( 2) Creation Date
	uint16_t fowner;		// ( 2) Owner ID
	uint16_t farights;		// ( 2) Access Rights
	uint16_t fmtime;		// ( 2) File Modification Time (MAKE_FILE_TIME)
	uint16_t fmdate;		// ( 2) File Modification Time (MAKE_FILE_DATE)
	uint16_t fscluster;		// ( 2) Start Cluster
	uint32_t fsize;			// ( 4) File Size
} dir;						// 32 Bytes in total


file *__COF = NULL;			// Currently Opened File
dir __COD;					// Currently Opened Directory
char __CWD[256] = "/";		// Current Working Directory

bool driverInit(void);										// Initalizes the Driver, returns true if successful

file *fileOpen(char *path, uint8_t mode);					// Opens the File specified by "path"
void fileClose(file *fp);									// Closes the File specified by a File object
uint8_t fileDelete(char *path);								// Deletes the File specified by "path" (File has to be closed first)
uint8_t fileRename(char *path, char *newName);				// Renames the File specified by "path" to "newName"
#define fileError(fp)				(fp->ferror)			// Returns the last error specific to a File

void fileRead(file *fp, char *buf, uint16_t count);			// Read "count" bytes from a File into "buf"
void fileWrite(file *fp, char *buf, uint16_t count);		// Write "count" bytes to a File from "buf"

void fileSeek(file *fp, uint32_t position);					// Move the File pointer to "position"
#define fileTell(fp)				(fp->fptr)				// Read out the current File pointer
#define fileSize(fp)				(fp->fsize)				// Read out the File's size

uint8_t dirDelete(char *path);								// Delete a specified Directory (fails if the Directory is not empty)
void dirChange(char *path, bool create);					// Change the __CWD__ (and create a new Directory if the specified one doesn't exist)
#define dirOpen(path)				openPath(path, 0, 0)
uint8_t dirRead();											// Reads out a File's data from the current directory, call multiple times to go through all files

// Reading the Creation/Modified Date/Time of a File in the current Directory
// "creation" == 1 means it reads the creation date/time instead of the last modified

// Date Format:
// 15 14 13 12 11 10 09 | 08 07 06 05 | 04 03 02 01 00
// Y  Y  Y  Y  Y  Y  Y  | M  M  M  M  | D  D  D  D  D

// Time Format:
// 15 14 13 12 11 | 10 09 08 07 06 05 | 04 03 02 01 00
// H  H  H  H  H  | M  M  M  M  M  M  | S  S  S  S  S

#define dirGetYear(creation)		(((((creation) ? (__COD.fcdate ) : (__COD.fmdate)) >>  9U) & 0x7F) + 1980)
#define dirGetMonth(creation)		((((creation) ? (__COD.fcdate ) : (__COD.fmdate)) >>  5U) & 0x0F)
#define dirGetDay(creation)			 (((creation) ? (__COD.fcdate ) : (__COD.fmdate))         & 0x1F)
#define dirGetHour(creation)		((((creation) ? (__COD.fctimes) : (__COD.fmtime)) >> 11U) & 0x1F)
#define dirGetMinute(creation)		((((creation) ? (__COD.fctimes) : (__COD.fmtime)) >>  5U) & 0x3F)
#define dirGetSecond(creation)		((((creation) ? (__COD.fctimes) : (__COD.fmtime)) <<  1U) & 0x3E) // Seconds multiplied by 2 because fuck







// ----------- //
// RTC Section //
// ----------- //


/* MCP795 Instructions, see datasheet table 3-1 */
#define MCP795_EEREAD			0x03
#define MCP795_EEWRITE			0x02
#define MCP795_EEWRDI			0x04
#define MCP795_EEWREN			0x06
#define MCP795_SRREAD			0x05
#define MCP795_SRWRITE			0x01
#define MCP795_READ				0x13
#define MCP795_WRITE			0x12
#define MCP795_UNLOCK			0x14
#define MCP795_IDWRITE			0x32
#define MCP795_IDREAD			0x33
#define MCP795_CLRWDT			0x44
#define MCP795_CLRRAM			0x54

/* MCP795 RTCC registers, see datasheet table 4-1 */
#define MCP795_REG_SECONDS		0x01
#define MCP795_REG_HOUR			0x03
#define MCP795_REG_DAY			0x04
#define MCP795_REG_MONTH		0x06
#define MCP795_REG_CONTROL		0x08
#define MCP795_REG_ALM0_SECONDS	0x0C
#define MCP795_REG_ALM0_DAY		0x0F

#define MCP795_ST_BIT			7U	// 0b10000000
#define MCP795_24_BIT			6U	// 0b01000000
#define MCP795_LP_BIT			5U	// 0b00100000
#define MCP795_EXTOSC_BIT		3U	// 0b00001000
#define MCP795_OSCON_BIT		5U	// 0b00100000
#define MCP795_VBATEN_BIT		3U	// 0b00001000
#define MCP795_ALM0_BIT			4U	// 0b00010000
#define MCP795_ALM1_BIT			5U	// 0b00100000
#define MCP795_ALM0IF_BIT		3U	// 0b00001000
#define MCP795_ALM0C0_BIT		4U	// 0b00010000
#define MCP795_ALM0C1_BIT		5U	// 0b00100000
#define MCP795_ALM0C2_BIT		6U	// 0b01000000

#define SEC_PER_DAY				(24 * 60 * 60)

#define bcd2bin(val)			(((val) & 0x0F) + ((val) >> 4U) * 10U)
#define bin2bcd(val)			((((val) / 10U) << 4U) + (val) % 10U)



typedef struct __rtc_time_struct{
	uint8_t tm_sec;		// Seconds		(0-59)
	uint8_t tm_min;		// Minutes		(0-59)
	uint8_t tm_hour;	// Hours		(0-23)
	uint8_t tm_wday;	// Week Day		(1-7 <-> monday-sunday)
	uint8_t tm_mday;	// Month Day	(1-31)
	uint8_t tm_mon;		// Month		(1-12)
	uint8_t tm_year;	// Year			(0-99)
} rtc_time;

void readRTC(uint8_t addr, uint8_t *buf, uint8_t count);
void writeRTC(uint8_t addr, uint8_t *buf, uint8_t count);

void setBitRTC(uint8_t addr, uint8_t bit, bool val);

bool stopRTC(bool *extosc);			// Returns false on failure, true on success
bool startRTC(bool *extosc);		// Returns false on failure, true on success

void readTime(rtc_time *tm);
void writeTime(rtc_time *tm);


